/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*******************************************************************************
* MPIA, NEVEC - MIDI project
*
* "@(#) $Id: array_geometry.c,v 1.5 2002/11/07 12:50:39 jaffe Exp $"
*
* who        when      what
* ---------  --------  ----------------------------------------------
* jdejong    2002-05-08 imported
*
*/
#define _POSIX_SOURCE_1
#include "vltPort.h"

/*----------------------------------------------------------------------- */
/* Array Geometry table FITS binary tables routines                       */
/*----------------------------------------------------------------------- */
#include "array_geometry.h"

/* internal prototypes */
/* get data from table_def specific to array_geometry_table table */
void get_array_geometry_table_info (array_geometry_table_def *table);

/* file global variables for table type*/
static char *tabletype = "ARRAY_GEOMETRY";


/* array_geometry_table_row functions */
array_geometry_table_row* 
make_array_geometry_table_row(array_geometry_table_def* table)
/*----------------------------------------------------------------------- */
/*   Make an array_geometry_table_row structure                           */
/*   Inputs:                                                              */
/*      table     pointer to array_geometry_table table structure         */
/*   Returns:                                                             */
/*      An array_geometry_table_row structure pointer ; NULL => failure   */
/*      Data pointers set to generic structure values                     */
/*----------------------------------------------------------------------- */
{
  int i;
  array_geometry_table_row *out;

  /* create basic structure */
  out = (array_geometry_table_row*)malloc(sizeof(array_geometry_table_row));
  if (out==NULL) return out; /* allocation failure? */
  
  /* zero memory */
  memset (out, 0, sizeof(array_geometry_table_row));

  /* Make generic row structure */
  out->raw = make_table_row_from_def(table->def);

  return out; /* done */
} /* end  make_array_geometry_table_row */

void kill_array_geometry_table_row (array_geometry_table_row *row)
/*----------------------------------------------------------------------- */
/*   Delete an array_geometry_table_row structure                         */
/*   Inputs:                                                              */
/*      row    pointer to the array_geometry_table_row structure.         */
/*----------------------------------------------------------------------- */
{
  if (row==NULL) return;         /* anybody home */
  kill_table_row(row->raw);      /* generic row structure */
  free (row);                              /* basic structure */
} /* end  kill_array_geometry_table_row */


/* array_geometry_table functions */

array_geometry_table_def* 
make_array_geometry_table_def(fitsfile *fptr, char *arrnam, char *date_obs,
			      double arrayxyz[3], int *ierr)
/*----------------------------------------------------------------------- */
/*   Make a new array_geometry_table in an open FITS file                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      arrnam    Array name                                              */
/*      date_obs  Date (yyyy-mm-dd) observations began                    */
/*      arrayxyz  Geodetic position of array center (m)                   */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      A array_geometry_table_def pointer; NULL indicates failure        */
/*----------------------------------------------------------------------- */
{
  array_geometry_table_def *out = NULL;
  int  version = 1; /* table version number */
  int  ikey, icol, i;
  char *frame = "GEOCENTRIC";

  if (*ierr>0) return NULL;    /* existing error condition? */

  if (fptr==NULL)     return NULL; /* sanity check */

  /* create output object */
  out = (array_geometry_table_def*)malloc(sizeof(array_geometry_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(array_geometry_table_def)); /* zero fill */

  /* create table_def */
  out->def = make_table_def(ARRAY_GEOMETRY_TABLE, 0, 0);
  if (out->def==NULL) { /* check for failure */
    kill_array_geometry_table_def(out); /* clean up */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating ARRAY_GEOMETRY_TABLE for %s table_def \n", 
	    tabletype);
    return NULL;
  }

  /* no mount parameter data */
  icol = table_find_col (out->def, "MNTPARAM")-1;
  if (icol>0) out->def->colcount[icol] = 0;

  /* add keyword values */
  ikey = table_find_key(out->def, "REVISION");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&version, ierr);
  ikey = table_find_key(out->def, "ARRNAM");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)arrnam, ierr);
  ikey = table_find_key(out->def, "DATE-OBS");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)date_obs, ierr);

  /* default is center of the earth */
  ikey = table_find_key(out->def, "FRAME");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)frame, ierr);

  ikey = table_find_key(out->def, "ARRAYX");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&arrayxyz[0], ierr);

  ikey = table_find_key(out->def, "ARRAYY");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&arrayxyz[1], ierr);

  ikey = table_find_key(out->def, "ARRAYZ");
  if ((ikey>0) && (*ierr==0)) 
    table_put_keyword (out->def, ikey, (void*)&arrayxyz[2], ierr);

  /* get info from table descriptor */
  get_array_geometry_table_info (out);

  /* open source table */
  make_table_from_def (fptr, out->def, ierr);

  return out; /* done */
} /* end make_array_geometry_table */

array_geometry_table_def* 
open_array_geometry_table(fitsfile *fptr, int *ierr)
/*----------------------------------------------------------------------- */
/*   Open an existing array_geometry_table in the active FITS file        */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Output:                                                              */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      An array_geometry_table_def pointer; NULL indicates failure       */
/*----------------------------------------------------------------------- */
{
  array_geometry_table_def *out = NULL;

  if (*ierr>0) return NULL;    /* existing error condition? */
  if (fptr==NULL) return NULL; /* sanity check */

  /* create output object */
  out = (array_geometry_table_def*)malloc(sizeof(array_geometry_table_def));
  if (out==NULL)     {  /* check for allocation failure */
    *ierr = 10000;
    fprintf(stderr,"ERROR creating %s table_def \n", tabletype);
    return out;
  }
  memset (out, 0, sizeof(array_geometry_table_def)); /* zero fill */

  /* position FITS file at  table */
  if (fits_movnam_hdu(fptr, BINARY_TBL, tabletype, 0, ierr)) {
    fprintf(stderr,"ERROR %d finding %s table \n", *ierr, tabletype);
    kill_array_geometry_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get table descriptor */
  out->def = make_table_def_from_table (fptr);
  if (out->def==NULL) { /* error check */
    fprintf(stderr,"ERROR making table def for %s \n", tabletype);
    kill_array_geometry_table_def(out); /* clean up */
    *ierr = 10000;
    return NULL;
  }

  /* get info from table descriptor */
  get_array_geometry_table_info (out);

  return out; /* done */
} /* end open_array_geometry_table */

int read_array_geometry_table(fitsfile *fptr, array_geometry_table_def* table, 
			    array_geometry_table_row *row, long row_number,
			    int *ierr)
/*----------------------------------------------------------------------- */
/*   Read the next row from an array_geometry_table                       */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row_number which row (1-rel) in the table                         */   
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      row       pointer to an extant row structure                      */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* read row */
  table_read_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  read_array_geometry_table: reading table %s \n", 
	    tabletype);
    return 1;
  }

  /* copy relevant data from generic table_row */
  trow = row->raw;
  strncpy (row->sta_name, trow->data[table->sta_name_col]->tdata.tchar, 8);
  strncpy (row->tel_name, trow->data[table->tel_name_col]->tdata.tchar, 8);
  row->index       = trow->data[table->index_col]->tdata.tshort[0];
  row->diameter    = trow->data[table->diameter_col]->tdata.tfloat[0];
  row->staxyz[0]   = trow->data[table->staxyz_col]->tdata.tdouble[0];
  row->staxyz[1]   = trow->data[table->staxyz_col]->tdata.tdouble[1];
  row->staxyz[2]   = trow->data[table->staxyz_col]->tdata.tdouble[2];
  row->mntsta      = trow->data[table->mntsta_col]->tdata.tshort[0];
  row->mntparam    = trow->data[table->mntparam_col]->tdata.tdouble;

  return 0; /* OK - done */
} /* end read_sourcetable */

int write_array_geometry_table(fitsfile *fptr, array_geometry_table_def* table, 
			     array_geometry_table_row *row, long row_number, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Write  the next row to an array_geometry_table                               */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*      table     table definitions structure from open or create         */
/*      row       pointer to an extant row structure                      */
/*      row_number which row (1-rel) in the table                         */
/*                 -1 => next row                                         */  
/*   Input/output:                                                        */
/*      ierr      Error code: 0 => ok, checked on input                   */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  table_row *trow;

  if (*ierr>0) return 1;     /* existing error condition? */
  if (fptr==NULL) return 1;  /* sanity check */
  if (table==NULL) return 1; /* sanity check */
  if (row==NULL) return 1;   /* sanity check */

  /* set row number */
  if (row_number>0) { /* explicit row number */
    row->raw->trow = row_number;
    table->row = row_number;
  } else { /* next row */
    row->raw->trow = ++table->row;
  }

  /* copy relevant data to generic table_row */
  trow = row->raw;
  strncpy (trow->data[table->sta_name_col]->tdata.tchar, row->sta_name, 8);
  strncpy (trow->data[table->tel_name_col]->tdata.tchar, row->tel_name, 8);
  trow->data[table->index_col]->tdata.tshort[0]     = row->index;
  trow->data[table->diameter_col]->tdata.tfloat[0]  = row->diameter;
  trow->data[table->staxyz_col]->tdata.tdouble[0]   = row->staxyz[0];
  trow->data[table->staxyz_col]->tdata.tdouble[1]   = row->staxyz[1];
  trow->data[table->staxyz_col]->tdata.tdouble[2]   = row->staxyz[2];
  trow->data[table->mntsta_col]->tdata.tshort[0]    = row->mntsta;
  trow->data[table->mntparam_col]->tdata.tdouble    = row->mntparam;

  /* write row */
  table_write_row(fptr, table->def, row->raw, ierr);
  if (*ierr>0) { /* error check */
    fprintf(stderr,"ERROR  write_array_geometry_table: writing table %s \n", 
	    tabletype);
    return 1;
  }

  return 0; /* OK - done */
} /* end write_array_geometry_table */

int close_array_geometry_table(fitsfile *fptr, array_geometry_table_def* table, 
			     int *ierr)
/*----------------------------------------------------------------------- */
/*   Close an open array_geometry_table, associated structures are        */
/*   destroyed                                                            */
/*   Note: this closes the full FITS file                                 */
/*   Inputs:                                                              */
/*      fptr      cfitsio pointer to open FITS file                       */
/*   Returns:                                                             */
/*      0 if successful, else failed                                      */
/*----------------------------------------------------------------------- */
{
  if (*ierr>0) return 1;    /* existing error condition? */
  if (fptr==NULL) return 1; /* sanity check */

  /* do actual file close */
  fits_close_file (fptr, ierr);

  /* delete structures */
  kill_array_geometry_table_def(table);

  return 0; /* OK - done */
} /* end close_array_geometry_table */

void kill_array_geometry_table_def (array_geometry_table_def* table)
/*----------------------------------------------------------------------- */
/*   Delete an array_geometry_table structure                             */
/*   Inputs:                                                              */
/*      table     table definitions structure from open or create         */
/*----------------------------------------------------------------------- */
{
  if (table==NULL) return; /* sanity check */

  /* table definition structure */
  kill_table_def(table->def);

  /* free basic structure */
  free (table);
} /* end kill_array_geometry_table_def */

/*************************private functions ******************************/

void get_array_geometry_table_info (array_geometry_table_def *table)
/*----------------------------------------------------------------------- */
/*   Get data from table_def specific to array_geometry_table table       */
/*   Uses detector member which must already be attached                  */
/*   Inputs:                                                              */
/*      table     table definitions structure                             */
/*----------------------------------------------------------------------- */
{
  int ikey, ierr=0;
  if (table==NULL) return;             /* sanity check */

  /* keyword values */
  ikey = table_find_key(table->def, "ARRNAM");
  if (ikey>0) table->arrnam = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "DATE-OBS");
  if (ikey>0) table->date_obs = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "FRAME");
  if (ikey>0) table->frame = (char*)table->def->keydata[ikey-1];
  ikey = table_find_key(table->def, "ARRAYX");
  if (ikey>0) table_get_keyword(table->def, ikey, (void*)&table->arrayx, 
				&ierr);
  ikey = table_find_key(table->def, "ARRAYY");
  if (ikey>0) table_get_keyword(table->def, ikey, (void*)&table->arrayy, 
				&ierr);
  ikey = table_find_key(table->def, "ARRAYZ");
  if (ikey>0) table_get_keyword(table->def, ikey, (void*)&table->arrayz, 
				&ierr);

  /* lookup column numbers (0-rel) */
  table->sta_name_col = table_find_col (table->def, "STA_NAME")-1;
  table->tel_name_col = table_find_col (table->def, "TEL_NAME")-1;
  table->index_col    = table_find_col (table->def, "INDEX")-1;
  table->diameter_col = table_find_col (table->def, "DIAMETER")-1;
  table->staxyz_col   = table_find_col (table->def, "STAXYZ")-1;
  table->mntsta_col   = table_find_col (table->def, "MNTSTA")-1;
  table->mntparam_col = table_find_col (table->def, "MNTPARAM")-1;

} /* end  get_array_geometry_table_info */

/* */
